@@ -2,19 +2,20 @@
 # this repository contains the full copyright notices and license terms.
 
 from collections import defaultdict
+from decimal import Decimal
 
+from trytond.i18n import gettext
 from trytond.model import ModelSQL, ModelView, Workflow, fields
+from trytond.model import DeactivableMixin
+from trytond.model.exceptions import AccessError
 from trytond.pyson import Eval, If, Bool
 from trytond.pool import Pool
 from trytond.transaction import Transaction
 
 from trytond.modules.product import price_digits
 
-__all__ = ['Type', 'Complaint', 'Action',
-    'Action_SaleLine', 'Action_InvoiceLine']
 
-
-class Type(ModelSQL, ModelView):
+class Type(DeactivableMixin, ModelSQL, ModelView):
     'Customer Complaint Type'
     __name__ = 'sale.complaint.type'
 
@@ -37,8 +38,12 @@ class Complaint(Workflow, ModelSQL, ModelView):
     number = fields.Char('Number', readonly=True, select=True)
     reference = fields.Char('Reference', select=True)
     date = fields.Date('Date', states=_states, depends=_depends)
-    customer = fields.Many2One('party.party', 'Customer', required=True,
-        states=_states, depends=_depends)
+    customer = fields.Many2One(
+        'party.party', "Customer", required=True, states=_states,
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=_depends + ['company'])
     address = fields.Many2One('party.address', 'Address',
         domain=[('party', '=', Eval('customer'))],
         states=_states, depends=_depends + ['customer'])
@@ -79,19 +84,17 @@ class Complaint(Workflow, ModelSQL, ModelView):
     def __setup__(cls):
         super(Complaint, cls).__setup__()
         cls._order.insert(0, ('date', 'DESC'))
-        cls._error_messages.update({
-                'delete_draft': ('Complaint "%s" must be in draft '
-                    'to be deleted.'),
-                })
         cls._transitions |= set((
                 ('draft', 'waiting'),
                 ('waiting', 'draft'),
                 ('waiting', 'approved'),
                 ('waiting', 'rejected'),
                 ('approved', 'done'),
+                ('approved', 'draft'),
                 ('draft', 'cancelled'),
                 ('waiting', 'cancelled'),
                 ('done', 'draft'),
+                ('rejected', 'draft'),
                 ('cancelled', 'draft'),
                 ))
         cls._buttons.update({
@@ -152,23 +155,33 @@ class Complaint(Workflow, ModelSQL, ModelView):
     def _origin_domains(cls):
         return {
             'sale.sale': [
-                ('party', '=', Eval('customer')),
+                If(Eval('customer'),
+                    ('party', '=', Eval('customer')),
+                    ()),
                 ('company', '=', Eval('company')),
                 ('state', 'in', ['confirmed', 'processing', 'done']),
                 ],
             'sale.line': [
-                ('sale.party', '=', Eval('customer')),
+                ('type', '=', 'line'),
+                If(Eval('customer'),
+                    ('sale.party', '=', Eval('customer')),
+                    ()),
                 ('sale.company', '=', Eval('company')),
                 ('sale.state', 'in', ['confirmed', 'processing', 'done']),
                 ],
             'account.invoice': [
-                ('party', '=', Eval('customer')),
+                If(Eval('customer'),
+                    ('party', '=', Eval('customer')),
+                    ()),
                 ('company', '=', Eval('company')),
                 ('type', '=', 'out'),
                 ('state', 'in', ['posted', 'paid']),
                 ],
             'account.invoice.line': [
-                ('invoice.party', '=', Eval('customer')),
+                ('type', '=', 'line'),
+                If(Eval('customer'),
+                    ('invoice.party', '=', Eval('customer')),
+                    ()),
                 ('invoice.company', '=', Eval('company')),
                 ('invoice.type', '=', 'out'),
                 ('invoice.state', 'in', ['posted', 'paid']),
@@ -207,6 +220,23 @@ class Complaint(Workflow, ModelSQL, ModelView):
         else:
             return []
 
+    @fields.depends('origin', 'customer')
+    def on_change_origin(self):
+        pool = Pool()
+        Sale = pool.get('sale.sale')
+        SaleLine = pool.get('sale.line')
+        Invoice = pool.get('account.invoice')
+        InvoiceLine = pool.get('account.invoice.line')
+        if not self.customer and self.origin and self.origin.id >= 0:
+            if isinstance(self.origin, Sale):
+                self.customer = self.origin.party
+            elif isinstance(self.origin, SaleLine):
+                self.customer = self.origin.sale.party
+            elif isinstance(self.origin, Invoice):
+                self.customer = self.origin.party
+            elif isinstance(self.origin, InvoiceLine) and self.origin.invoice:
+                self.customer = self.origin.invoice.party
+
     @fields.depends('origin')
     def on_change_with_origin_id(self, name=None):
         if self.origin:
@@ -217,18 +247,25 @@ class Complaint(Workflow, ModelSQL, ModelView):
         if self.origin:
             return self.origin.__class__.__name__
 
+    @classmethod
+    def view_attributes(cls):
+        return super().view_attributes() + [
+            ('/tree', 'visual', If(Eval('state') == 'cancelled', 'muted', '')),
+            ]
+
     @classmethod
     def create(cls, vlist):
         pool = Pool()
-        Sequence = pool.get('ir.sequence')
         Configuration = pool.get('sale.configuration')
 
         config = Configuration(1)
         vlist = [v.copy() for v in vlist]
+        default_company = cls.default_company()
         for values in vlist:
             if values.get('number') is None:
-                values['number'] = Sequence.get_id(
-                    config.complaint_sequence.id)
+                values['number'] = config.get_multivalue(
+                    'complaint_sequence',
+                    company=values.get('company', default_company)).get()
         return super(Complaint, cls).create(vlist)
 
     @classmethod
@@ -244,7 +281,9 @@ class Complaint(Workflow, ModelSQL, ModelView):
     def delete(cls, complaints):
         for complaint in complaints:
             if complaint.state != 'draft':
-                cls.raise_user_error('delete_draft', complaint.rec_name)
+                raise AccessError(
+                    gettext('sale_complaint.msg_complaint_delete_draft',
+                        complaint=complaint.rec_name))
         super(Complaint, cls).delete(complaints)
 
     @classmethod
@@ -269,7 +308,13 @@ class Complaint(Workflow, ModelSQL, ModelView):
     @ModelView.button
     @Workflow.transition('approved')
     def approve(cls, complaints):
-        pass
+        pool = Pool()
+        Configuration = pool.get('sale.configuration')
+        config = Configuration(1)
+        with Transaction().set_context(
+                queue_name='sale',
+                queue_scheduled_at=config.sale_process_after):
+            cls.__queue__.process(complaints)
 
     @classmethod
     @ModelView.button
@@ -304,9 +349,10 @@ class Action(ModelSQL, ModelView):
     __name__ = 'sale.complaint.action'
 
     _states = {
-        'readonly': Bool(Eval('result')),
+        'readonly': ((Eval('complaint_state') != 'draft')
+            | Bool(Eval('result'))),
         }
-    _depends = ['result']
+    _depends = ['complaint_state', 'result']
     _line_states = {
         'invisible': ~Eval('_parent_complaint', {}
             ).get('origin_model', 'sale.line').in_(
@@ -315,29 +361,27 @@ class Action(ModelSQL, ModelView):
         }
     _line_depends = _depends
 
-    complaint = fields.Many2One('sale.complaint', 'Complaint', required=True,
+    complaint = fields.Many2One(
+        'sale.complaint', 'Complaint', required=True, ondelete='CASCADE',
         states=_states, depends=_depends)
     action = fields.Selection([
             ('sale_return', 'Create Sale Return'),
             ('credit_note', 'Create Credit Note'),
-            ], 'Action')
+            ], 'Action', states=_states)
 
-    sale_lines = fields.Many2Many('sale.complaint.action-sale.line',
-        'action', 'line', 'Sale Lines',
-        domain=[('sale', '=', Eval('_parent_complaint', {}).get('origin_id'))],
+    sale_lines = fields.One2Many(
+        'sale.complaint.action-sale.line', 'action', "Sale Lines",
         states={
             'invisible': Eval('_parent_complaint', {}
                 ).get('origin_model', 'sale.sale') != 'sale.sale',
             'readonly': _states['readonly'],
             },
         depends=_depends,
-        help='Leave empty for all lines')
+        help='Leave empty for all lines.')
 
-    invoice_lines = fields.Many2Many(
-        'sale.complaint.action-account.invoice.line', 'action', 'line',
-        'Invoice Lines',
-        domain=[('invoice', '=', Eval('_parent_complaint', {}
-                    ).get('origin_id'))],
+    invoice_lines = fields.One2Many(
+        'sale.complaint.action-account.invoice.line', 'action',
+        "Invoice Lines",
         states={
             'invisible': Eval('_parent_complaint', {}
                 ).get('origin_model', 'account.invoice.line'
@@ -345,30 +389,41 @@ class Action(ModelSQL, ModelView):
             'readonly': _states['readonly'],
             },
         depends=_depends,
-        help='Leave empty for all lines')
+        help='Leave empty for all lines.')
 
     quantity = fields.Float('Quantity',
         digits=(16, Eval('unit_digits', 2)),
         states=_line_states, depends=_line_depends + ['unit_digits'],
-        help='Leave empty for the same quantity')
+        help='Leave empty for the same quantity.')
     unit = fields.Function(fields.Many2One('product.uom', 'Unit',
             states=_line_states, depends=_line_depends),
         'on_change_with_unit')
-    unit_digits = fields.Function(fields.Integer('Unit Digits'),
-        'get_unit_digits')
+    unit_digits = fields.Function(
+        fields.Integer('Unit Digits'), 'on_change_with_unit_digits')
     unit_price = fields.Numeric('Unit Price', digits=price_digits,
         states=_line_states, depends=_line_depends,
-        help='Leave empty for the same price')
+        help='Leave empty for the same price.')
+
+    amount = fields.Function(fields.Numeric(
+            "Amount", digits=(16, Eval('currency_digits', 2)),
+            depends=['currency_digits']),
+        'on_change_with_amount')
+    currency = fields.Function(fields.Many2One(
+            'currency.currency', "Currency"),
+        'on_change_with_currency')
+    currency_digits = fields.Function(fields.Integer("Currency Digits"),
+        'on_change_with_currency_digits')
 
     result = fields.Reference('Result', selection='get_result', readonly=True)
 
+    complaint_state = fields.Function(
+        fields.Selection('get_complaint_states', "Complaint State"),
+        'on_change_with_complaint_state')
+
     @classmethod
     def __setup__(cls):
-        super(Action, cls).__setup__()
-        cls._error_messages.update({
-                'delete_result': ('Action "%s" must not have result '
-                    'to be deleted.'),
-                })
+        super().__setup__()
+        cls.__access__.add('complaint')
 
     @fields.depends('complaint',
         '_parent_complaint.origin_model', '_parent_complaint.origin')
@@ -378,13 +433,82 @@ class Action(ModelSQL, ModelView):
                     'sale.line', 'account.invoice.line'}):
             return self.complaint.origin.unit.id
 
-    @fields.depends('complaint')
-    def get_unit_digits(self, name=None):
+    @fields.depends('complaint',
+        '_parent_complaint.origin_model', '_parent_complaint.origin')
+    def on_change_with_unit_digits(self, name=None):
         if (self.complaint
                 and self.complaint.origin_model in {
                     'sale.line', 'account.invoice.line'}):
             return self.complaint.origin.unit.digits
 
+    @fields.depends(
+        'quantity', 'unit_price', 'currency', 'sale_lines', 'invoice_lines',
+        'complaint', '_parent_complaint.origin_model',
+        '_parent_complaint.origin')
+    def on_change_with_amount(self, name=None):
+        if self.complaint:
+            if self.complaint.origin_model in {
+                    'sale.line', 'account.invoice.line'}:
+                if self.quantity is not None:
+                    quantity = self.quantity
+                else:
+                    quantity = self.complaint.origin.quantity
+                if self.unit_price is not None:
+                    unit_price = self.unit_price
+                else:
+                    unit_price = self.complaint.origin.unit_price
+                amount = Decimal(str(quantity)) * unit_price
+                if self.currency:
+                    amount = self.currency.round(amount)
+                return amount
+            elif self.complaint.origin_model == 'sale.sale':
+                if not self.sale_lines:
+                    if self.complaint and self.complaint.origin:
+                        return self.complaint.origin.untaxed_amount
+                else:
+                    return sum(
+                        getattr(l, 'amount', None) or Decimal(0)
+                        for l in self.sale_lines)
+            elif self.complaint.origin_model == 'account.invoice':
+                if not self.invoice_lines:
+                    if self.complaint and self.complaint.origin:
+                        return self.complaint.origin.untaxed_amount
+                else:
+                    return sum(
+                        getattr(l, 'amount', None) or Decimal(0)
+                        for l in self.invoice_lines)
+
+    @fields.depends(
+        'complaint',
+        '_parent_complaint.origin_model', '_parent_complaint.origin')
+    def on_change_with_currency(self, name=None):
+        if (self.complaint
+                and self.complaint.origin_model in {
+                    'sale.sale', 'sale.line',
+                    'account.invoice', 'account.invoice.line'}):
+            return self.complaint.origin.currency.id
+
+    @fields.depends(
+        'complaint',
+        '_parent_complaint.origin_model', '_parent_complaint.origin')
+    def on_change_with_currency_digits(self, name=None):
+        if (self.complaint
+                and self.complaint.origin_model in {
+                    'sale.sale', 'sale.line',
+                    'account.invoice', 'account.invoice.line'}):
+            return self.complaint.origin.currency.digits
+
+    @classmethod
+    def get_complaint_states(cls):
+        pool = Pool()
+        Complaint = pool.get('sale.complaint')
+        return Complaint.fields_get(['state'])['state']['selection']
+
+    @fields.depends('complaint', '_parent_complaint.state')
+    def on_change_with_complaint_state(self, name=None):
+        if self.complaint:
+            return self.complaint.state
+
     @classmethod
     def _get_result(cls):
         'Return list of Model names for result Reference'
@@ -394,11 +518,9 @@ class Action(ModelSQL, ModelView):
     def get_result(cls):
         pool = Pool()
         Model = pool.get('ir.model')
+        get_name = Model.get_name
         models = cls._get_result()
-        models = Model.search([
-                ('model', 'in', models),
-                ])
-        return [(None, '')] + [(m.model, m.name) for m in models]
+        return [(None, '')] + [(m, get_name(m)) for m in models]
 
     @classmethod
     def copy(cls, actions, default=None):
@@ -421,7 +543,16 @@ class Action(ModelSQL, ModelView):
             default = {}
             if isinstance(self.complaint.origin, Sale):
                 sale = self.complaint.origin
-                sale_lines = self.sale_lines or sale.lines
+                if self.sale_lines:
+                    sale_lines = [l.line for l in self.sale_lines]
+                    line2qty = {
+                        l.line.id: l.get_quantity() for l in self.sale_lines}
+                    line2price = {
+                        l.line.id: l.get_unit_price() for l in self.sale_lines}
+                    default['quantity'] = lambda o: line2qty.get(o['id'])
+                    default['unit_price'] = lambda o: line2price.get(o['id'])
+                else:
+                    sale_lines = [l for l in sale.lines if l.type == 'line']
             elif isinstance(self.complaint.origin, Line):
                 sale_line = self.complaint.origin
                 sale = sale_line.sale
@@ -448,30 +579,50 @@ class Action(ModelSQL, ModelView):
         Line = pool.get('account.invoice.line')
 
         if isinstance(self.complaint.origin, (Invoice, Line)):
+            line2qty = line2price = {}
             if isinstance(self.complaint.origin, Invoice):
                 invoice = self.complaint.origin
-                invoice_lines = self.invoice_lines or invoice.lines
+                if self.invoice_lines:
+                    invoice_lines = [l.line for l in self.invoice_lines]
+                    line2qty = {l.line: l.quantity
+                        for l in self.invoice_lines
+                        if l.quantity is not None}
+                    line2price = {l.line: l.unit_price
+                        for l in self.invoice_lines
+                        if l.unit_price is not None}
+                else:
+                    invoice_lines = [
+                        l for l in invoice.lines if l.type == 'line']
             elif isinstance(self.complaint.origin, Line):
                 invoice_line = self.complaint.origin
                 invoice = invoice_line.invoice
                 invoice_lines = [invoice_line]
-            credit_note = invoice._credit()
-            credit_lines = []
-            for invoice_line in invoice_lines:
-                credit_line = invoice_line._credit()
-                credit_lines.append(credit_line)
-                # Remove product as it is not a return
-                credit_line.product = None
-                credit_line.origin = self.complaint
-            if isinstance(self.complaint.origin, Line):
                 if self.quantity is not None:
-                    credit_lines[0].quantity = -self.quantity
+                    line2qty = {invoice_line: self.quantity}
                 if self.unit_price is not None:
-                    credit_lines[0].unit_price = self.unit_price
-            credit_note.lines = credit_lines
-            credit_note.taxes = None
-            credit_note.save()
-            Invoice.update_taxes([credit_note])
+                    line2price = {invoice_line: self.unit_price}
+            with Transaction().set_context(_account_invoice_correction=True):
+                credit_note, = Invoice.copy([invoice], default={
+                        'lines': [],
+                        })
+                # Copy each line one by one to get negative and positive lines
+                # following each other
+                for invoice_line in invoice_lines:
+                    qty = line2qty.get(invoice_line, invoice_line.quantity)
+                    unit_price = invoice_line.unit_price - line2price.get(
+                        invoice_line, invoice_line.unit_price)
+                    Line.copy([invoice_line], default={
+                            'invoice': credit_note.id,
+                            'quantity': -qty,
+                            'origin': str(self.complaint),
+                            })
+                    credit_line, = Line.copy([invoice_line], default={
+                            'invoice': credit_note.id,
+                            'quantity': qty,
+                            'unit_price': unit_price,
+                            'origin': str(self.complaint),
+                            })
+            credit_note.update_taxes()
         else:
             return
         return credit_note
@@ -480,25 +631,179 @@ class Action(ModelSQL, ModelView):
     def delete(cls, actions):
         for action in actions:
             if action.result:
-                cls.raise_user_error('delete_result', action.rec_name)
+                raise AccessError(
+                    gettext('sale_complaint.msg_action_delete_result',
+                        action=action.rec_name))
         super(Action, cls).delete(actions)
 
 
-class Action_SaleLine(ModelSQL):
+class _Action_Line:
+
+    _states = {
+        'readonly': (
+            (Eval('complaint_state') != 'draft')
+            | Bool(Eval('_parent_action.result', True))),
+        }
+    _depends = ['complaint_state']
+
+    action = fields.Many2One('sale.complaint.action', 'Action',
+        ondelete='CASCADE', select=True, required=True)
+    quantity = fields.Float(
+        "Quantity",
+        digits=(16, Eval('unit_digits', 2)),
+        states=_states,
+        depends=_depends + ['unit_digits'])
+    unit = fields.Function(
+        fields.Many2One('product.uom', "Unit"), 'on_change_with_unit')
+    unit_digits = fields.Function(
+        fields.Integer("Unit Digits"), 'on_change_with_unit_digits')
+    unit_price = fields.Numeric(
+        "Unit Price", digits=price_digits, states=_states, depends=_depends,
+        help='Leave empty for the same price.')
+
+    amount = fields.Function(fields.Numeric(
+            "Amount", digits=(16, Eval('currency_digits', 2)),
+            depends=['currency_digits']),
+        'on_change_with_amount')
+    currency = fields.Function(fields.Many2One(
+            'currency.currency', "Currency"),
+        'on_change_with_currency')
+    currency_digits = fields.Function(fields.Integer("Currency Digits"),
+        'on_change_with_currency_digits')
+
+    complaint_state = fields.Function(
+        fields.Selection('get_complaint_states', "Complaint State"),
+        'on_change_with_complaint_state')
+    complaint_origin_id = fields.Function(
+        fields.Integer("Complaint Origin ID"),
+        'on_change_with_complaint_origin_id')
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.add('action')
+
+    def on_change_with_unit(self, name=None):
+        raise NotImplementedError
+
+    def on_change_with_unit_digits(self, name=None):
+        raise NotImplementedError
+
+    @fields.depends('currency', methods=['get_quantity', 'get_unit_price'])
+    def on_change_with_amount(self, name=None):
+        quantity = self.get_quantity() or 0
+        unit_price = self.get_unit_price() or Decimal(0)
+        amount = Decimal(str(quantity)) * unit_price
+        if self.currency:
+            amount = self.currency.round(amount)
+        return amount
+
+    def get_quantity(self):
+        raise NotImplementedError
+
+    def get_unit_price(self):
+        raise NotImplementedError
+
+    @fields.depends('action', '_parent_action.currency')
+    def on_change_with_currency(self, name=None):
+        if self.action and self.action.currency:
+            return self.action.currency.id
+
+    @fields.depends('action', '_parent_action.currency')
+    def on_change_with_currency_digits(self, name=None):
+        if self.action and self.action.currency:
+            return self.action.currency.digits
+
+    @classmethod
+    def get_complaint_states(cls):
+        pool = Pool()
+        Complaint = pool.get('sale.complaint')
+        return Complaint.fields_get(['state'])['state']['selection']
+
+    @fields.depends('action', '_parent_action.complaint',
+        '_parent_action._parent_complaint.state')
+    def on_change_with_complaint_state(self, name=None):
+        if self.action and self.action.complaint:
+            return self.action.complaint.state
+
+    @fields.depends('action', '_parent_action.complaint',
+        '_parent_action._parent_complaint.origin_id')
+    def on_change_with_complaint_origin_id(self, name=None):
+        if self.action and self.action.complaint:
+            return self.action.complaint.origin_id
+
+
+class Action_SaleLine(_Action_Line, ModelView, ModelSQL):
     'Customer Complaint Action - Sale Line'
     __name__ = 'sale.complaint.action-sale.line'
 
-    action = fields.Many2One('sale.complaint.action', 'Action',
-        ondelete='CASCADE', select=True, required=True)
-    line = fields.Many2One('sale.line', 'Sale Line', ondelete='RESTRICT',
-        required=True)
+    line = fields.Many2One(
+        'sale.line', "Sale Line",
+        ondelete='RESTRICT', required=True,
+        domain=[
+            ('type', '=', 'line'),
+            ('sale', '=', Eval('complaint_origin_id', -1)),
+            ],
+        depends=['complaint_origin_id'])
+
+    @fields.depends('line')
+    def on_change_with_unit(self, name=None):
+        if self.line:
+            return self.line.unit.id
+
+    @fields.depends('line')
+    def on_change_with_unit_digits(self, name=None):
+        if self.line:
+            return self.line.unit.digits
+
+    @fields.depends('quantity', 'line')
+    def get_quantity(self):
+        if self.quantity is not None:
+            return self.quantity
+        elif self.line:
+            return self.line.quantity
+
+    @fields.depends('unit_price', 'line')
+    def get_unit_price(self):
+        if self.unit_price is not None:
+            return self.unit_price
+        elif self.line:
+            return self.line.unit_price
 
 
-class Action_InvoiceLine(ModelSQL):
+class Action_InvoiceLine(_Action_Line, ModelView, ModelSQL):
     'Customer Complaint Action - Invoice Line'
     __name__ = 'sale.complaint.action-account.invoice.line'
 
-    action = fields.Many2One('sale.complaint.action', 'Action',
-        ondelete='CASCADE', select=True, required=True)
-    line = fields.Many2One('account.invoice.line', 'Invoice Line',
-        ondelete='RESTRICT', required=True)
+    line = fields.Many2One(
+        'account.invoice.line', 'Invoice Line',
+        ondelete='RESTRICT', required=True,
+        domain=[
+            ('type', '=', 'line'),
+            ('invoice', '=', Eval('complaint_origin_id', -1)),
+            ],
+        depends=['complaint_origin_id'])
+
+    @fields.depends('line')
+    def on_change_with_unit(self, name=None):
+        if self.line:
+            return self.line.unit.id
+
+    @fields.depends('line')
+    def on_change_with_unit_digits(self, name=None):
+        if self.line:
+            return self.line.unit.digits
+
+    @fields.depends('quantity', 'line')
+    def get_quantity(self):
+        if self.quantity is not None:
+            return self.quantity
+        elif self.line:
+            return self.line.quantity
+
+    @fields.depends('unit_price', 'line')
+    def get_unit_price(self):
+        if self.unit_price is not None:
+            return self.unit_price
+        elif self.line:
+            return self.line.unit_price
